import datetime
import pdb

import pytest

from blackfynn import TimeSeries, TimeSeriesChannel
from blackfynn.models import TimeSeriesAnnotation, TimeSeriesAnnotationLayer


@pytest.fixture()
def timeseries(client, dataset):
    # create
    ts = TimeSeries('Human EEG')
    assert not ts.exists
    dataset.add(ts)
    assert ts.exists
    assert ts in dataset
    assert ts.type == 'TimeSeries'
    assert ts.name.startswith('Human EEG') #starts with bc duplicate names are appended
    ts2 = client.get(ts.id)
    assert ts2.id == ts.id
    assert ts2.name == ts.name
    assert ts2.type == 'TimeSeries'
    del ts2

    # provide to other tests
    yield ts

    # remove
    dataset.remove(ts)
    assert not ts.exists
    assert ts not in dataset


@pytest.fixture()
def timeseries2(client, dataset):
    ts = TimeSeries('Animal EEG')
    dataset.add(ts)
    assert ts.exists
    yield ts
    dataset.remove(ts)


def test_update_timeseries_name(client, timeseries):
    # update timeseries: change name
    timeseries.name = 'Monkey EEG'
    timeseries.update()
    timeseries2 = client.get(timeseries.id)
    assert timeseries2.id == timeseries.id
    assert timeseries2.name == timeseries.name
    assert timeseries2.type == 'TimeSeries'


def test_timeseries_channels(client, timeseries):
    num_channels = 10

    assert timeseries.exists

    ts2 = client.get(timeseries.id)

    chs = []
    for i in range(num_channels):
        # init
        chname = 'Channel-{}'.format(i)
        ch = TimeSeriesChannel(
                name=chname,
                rate=256,
                unit='uV')
        assert not ch.exists

        # create
        timeseries.add_channels(ch)
        assert ch.exists
        assert ch.name == chname
        chs.append(ch)

        # use separate request to get channel
        ch.insert_property('key','value')
        ch2 = [x for x in ts2.channels if x.id==ch.id][0]
        assert ch2.get_property('key') is not None
        assert ch2.name == ch.name
        assert ch2.type == ch.type
        assert ch2.rate == ch.rate
        assert ch2.unit == ch.unit
        del ch2

        # update channel
        ch.name = '{}-updated'.format(ch.name)
        ch.rate = 200
        ch.update()

        # use separate request to confirm name change
        ch2 = [x for x in ts2.channels if x.id==ch.id][0]
        assert ch2.name == ch.name
        assert ch2.rate == ch.rate

    # ensure correct number of channels
    channels = timeseries.channels
    assert len(channels) == num_channels

    # get channels (via API)
    ch_ids = [x.id for x in channels]

    # check
    for ch in chs:
        assert ch.id in ch_ids

    ch = channels[0]
    timeseries.remove_channels(ch)
    channels = timeseries.channels
    assert len(channels) == num_channels - 1
    assert ch not in channels
    assert ch.id not in [x.id for x in channels]

    # remove by id
    ch_id = channels[0].id
    timeseries.remove_channels(ch_id)
    channels = timeseries.channels
    assert len(channels) == num_channels - 2
    assert ch_id not in [x.id for x in channels]


def test_timeseries_annotations(client, timeseries):
    assert timeseries.exists
    print('layers = ', timeseries.layers)

    #Create Layer
    layer1 = TimeSeriesAnnotationLayer(name="test_layer", time_series_id = timeseries.id, description="test_description")
    a = layer1.as_dict()
    assert a['name'] == 'test_layer'
    assert a['description'] == 'test_description'

    # Add Layer
    timeseries.add_layer(layer1)
    assert layer1.exists

    # Get Layer
    layer1b = timeseries.get_layer('test_layer')
    assert layer1b.exists
    assert layer1b.name == "test_layer"
    assert layer1.id == layer1b.id
    assert layer1._api.timeseries is not None

    # Add another layer
    layer2 = timeseries.add_layer('test_layer2','test_description2')
    assert layer2.exists

    layer2copy = timeseries.add_layer('test_layer2')
    assert layer2copy.id == layer2.id

    # Get Layer
    layer2b = timeseries.get_layer('test_layer2')
    assert layer2b.exists
    assert layer2b.name == "test_layer2"
    assert layer2.id == layer2b.id
    assert layer2._api.timeseries is not None

    layers = timeseries.layers
    assert len(layers)==2
    assert layers[0].name == "test_layer"
    assert layers[1].name == 'test_layer2'

    #Create channels
    ch = TimeSeriesChannel(
        name='test_channel',
        rate=256,
        unit='uV',
        start=1,
        end=60*1e6)

    #Create annotation over one channel
    # create
    timeseries.add_channels(ch)
    assert ch.exists
    assert ch.id in [x.id for x in timeseries.channels]
    annot = TimeSeriesAnnotation(label = 'test_label', channel_ids = timeseries.channels[0].id,start = timeseries.channels[0].start,end=timeseries.channels[0].start+1*1e6)
    #Add Annotation
    layer1.add_annotations(annot)
    assert annot.exists

    # get annotations
    annotb = layer1.annotations()
    assert annotb[0].label == annot.label

    annotc = client._api.timeseries.get_annotation(timeseries, layer1, annot)
    assert annotc.label == annot.label

    #Create annotation over multiple channels
    ch2 = TimeSeriesChannel(
        name='test_channel',
        rate=256,
        unit='uV',
        start=1,
        end=60*1e6)

    timeseries.add_channels(ch2)
    channels = timeseries.channels
    ch_ids = [x.id for x in channels]
    assert ch2.exists
    assert ch2.id in ch_ids
    assert ch.id in ch_ids
    for ch in channels:
        assert ch.rate == 256
        assert ch.exists

    #add annotation over two channels
    channel_ids = [timeseries.channels[x].id for x in range(len(timeseries.channels))]

    annot2 = layer1.insert_annotation(annotation='test_label2', channel_ids = channel_ids,
        start =timeseries.channels[0].start+1*1e6,
        end = timeseries.channels[0].start+2*1e6)


    assert annot2.exists

    annot_gen = layer1.iter_annotations(1)
    annot = next(annot_gen)
    assert annot[0].label == 'test_label'
    next_annot= next(annot_gen)
    assert next_annot[0].label == 'test_label2'

    ### TEST ANNOTATION COUNTS
    layer1_expected_counts = [
        {'start': 0, 'end': 250000, 'value': 1.0},
        {'start': 250000, 'end': 500000, 'value': 1.0},
        {'start': 500000, 'end': 750000, 'value': 1.0},
        {'start': 750000, 'end': 1000000, 'value': 1.0},
        {'start': 1000000, 'end': 1250000, 'value': 2.0},
        {'start': 1250000, 'end': 1500000, 'value': 1.0},
        {'start': 1500000, 'end': 1750000, 'value': 1.0},
        {'start': 1750000, 'end': 2000000, 'value': 1.0},
        {'start': 2000000, 'end': 2000001, 'value': 1.0}
    ]

    def _sort_counts(counts):
        return sorted(counts, key=lambda c: c['start'])

    assert _sort_counts(
        timeseries.annotation_counts(
            start = timeseries.channels[0].start * 1e6,
            end = timeseries.channels[0].start + 2 * 1e6,
            layers=[layer1], period="0.25s"
        )[str(layer1.id)]
    ) == layer1_expected_counts
    assert _sort_counts(
        layer1.annotation_counts(
            start = timeseries.channels[0].start * 1e6,
            end = timeseries.channels[0].start + 2 * 1e6,
            period="0.25s"
        )[str(layer1.id)]
    ) == layer1_expected_counts

    ### TEST DELETION

    annot3 = TimeSeriesAnnotation(label= 'test_label3', channel_ids= channel_ids,start = timeseries.channels[0].start+1*1e6,end=timeseries.channels[0].start+2*1e6)
    layer1.add_annotations([annot2,annot3])
    annot3 = timeseries.add_annotations(layer=layer1,annotations=annot3)
    assert annot3.exists
    annot3.delete()
    assert not annot3.exists

    annot4 = timeseries.insert_annotation(layer=layer1,annotation='test_label3',start= timeseries.channels[0].start+1*1e6,end=timeseries.channels[0].start+2*1e6)
    assert annot4.exists
    annot4.delete()
    assert not annot4.exists

    annot5 = timeseries.insert_annotation(layer='test_layer4',annotation='test_label3',start=timeseries.channels[0].start+1*1e6,end=timeseries.channels[0].start+2*1e6)
    assert annot5.exists
    annot5.delete()
    assert not annot5.exists

    layer1.add_annotations([annot2,annot3])
    assert annot2.exists
    assert annot3.exists

    # test datetime input
    annot4 = timeseries.insert_annotation(layer='test_layer4',annotation='test_label3',start= datetime.datetime.utcfromtimestamp((timeseries.channels[0].start+1*1e6)/1e6),end=datetime.datetime.utcfromtimestamp(timeseries.channels[0].start+2*1e6))
    assert annot4.exists
    annot4.delete()
    assert not annot4.exists

    layer = timeseries.get_layer('test_layer4')
    assert layer.exists
    layer.delete()
    assert not layer.exists

    # delete annotations
    annot[0].delete()
    assert not annot[0].exists

    assert timeseries.exists
    timeseries.delete_layer(layer1)
    assert not layer1.exists

    assert layer2.exists
    layer2.delete()
    assert not layer2.exists


def test_timeseries_annotations_can_get_more_than_default_limit(timeseries):
    ch = TimeSeriesChannel(
        name='test_channel',
        rate=256,
        unit='uV',
        start=0,
        end=2e6)
    timeseries.add_channels(ch)

    layer = TimeSeriesAnnotationLayer(
        name="limit_layer", time_series_id=timeseries.id,
        description="layer with many small annotations")
    timeseries.add_layer(layer)

    for i in range(200):
        annotation = TimeSeriesAnnotation(
            label='annotation_{}'.format(i),
            channel_ids=ch.id,
            start=i*1e4,
            end=(i+1)*1e4)
        layer.add_annotations(annotation)

    # Should retrieve all annotations
    assert len(layer.annotations()) == 200

    # All of these annotations should be in the first 10-second window
    annots = next(layer.iter_annotations())
    assert len(annots) == 200

    layer.delete()


def test_timeseries_annotations_check_that_channels_belong_to_series(timeseries, timeseries2):
    ch = TimeSeriesChannel(
        name='test_channel',
        rate=256,
        unit='uV',
        start=0,
        end=1e6)
    timeseries.add_channels(ch)
    print(timeseries, timeseries2)

    ch2 = TimeSeriesChannel(
        name='test_channel_2',
        rate=256,
        unit='uV',
        start=0,
        end=1e6)
    timeseries2.add_channels(ch2)

    layer = TimeSeriesAnnotationLayer(
        name="a_layer", time_series_id=timeseries.id, description="a layer")
    timeseries.add_layer(layer)

    with pytest.raises(Exception):
        layer.annotations(channels=[ch, ch2])

    with pytest.raises(Exception):
        next(layer.iter_annotations(channels=[ch, ch2]))

    with pytest.raises(Exception):
        layer.annotation_counts(0, 1e6, period='0.1s', channels=[ch, ch2])


def test_timeseries_segments(client, timeseries):
    """
    No valid timeseries data to test against, but we can test the API calls.
    """
    channels = [
        TimeSeriesChannel(name='ch 1', rate=256,     unit='uV'),
        TimeSeriesChannel(name='ch 2', rate=111.123, unit='uV'),
    ]
    timeseries.add_channels(*channels)

    for ch in timeseries.channels:
        ch.end = ch.start + 5e6
        ch.update()

        ch.segments()
        ch.segments(start=ch.start+1e6)
        ch.segments(stop=ch.end-1e6)
        ch.segments(stop=ch.end-1e6, gap_factor=3)
        ch.segments(stop=ch.end-1e6, gap_factor=33.33)
        ch.segments(stop=ch.end-1e6, gap_factor=0.5)

    timeseries.segments()
    timeseries.segments(gap_factor=3)
    timeseries.segments(gap_factor=33.33)
    timeseries.segments(gap_factor=0.5)

    # TODO: API should soon error when value is non-numeric.
    #       Uncomment below when API is changed.

    with pytest.raises(Exception):
        timeseries.segments(gap_factor="should be int")

    with pytest.raises(Exception):
        ch.segments(gap_factor="should be int")



import datetime
import os

import pytest
import requests

from blackfynn import Blackfynn
from blackfynn.base import UnauthorizedException
from dateutil.parser import parse

# client library
from blackfynn.models import BaseNode, DataPackage, Dataset, File, PublishInfo, UserCollaborator, TeamCollaborator

from .utils import get_test_client


def test_basenode(client, dataset):
    node1 = dataset
    node2 = client.get_dataset(dataset.id)
    node3 = BaseNode()
    assert node1 == node2
    assert node1 != node3
    assert node1 != object()
    assert [node1, node2] == [node2, node1]

def test_update_dataset(client, dataset, session_id):
    # update name of dataset
    ds_name = 'Same Dataset, Different Name {}'.format(session_id)
    dataset.name = ds_name
    dataset.update()
    ds2 = client.get_dataset(dataset.id)
    assert ds2.id == dataset.id
    assert(isinstance(ds2.int_id, int))
    assert ds2.int_id == dataset.int_id
    assert ds2.name == ds_name
    assert ds2.owner_id == client.profile.id

def test_dataset_status_log(client, dataset):
    dataset_status_log = dataset.status_log()
    assert dataset_status_log.limit == 25
    assert dataset_status_log.offset == 0
    assert dataset_status_log.total_count == 1
    assert len(dataset_status_log.entries) == 1
    assert dataset_status_log.entries[0].user.node_id == client.profile.id
    assert dataset_status_log.entries[0].user.first_name == client.profile.first_name
    assert dataset_status_log.entries[0].user.last_name == client.profile.last_name
    assert dataset_status_log.entries[0].updated_at == parse(dataset.created_at)
    assert dataset_status_log.entries[0].status.id is not None
    assert dataset_status_log.entries[0].status.name is not None
    assert dataset_status_log.entries[0].status.display_name is not None

def test_status_is_readonly(client, dataset):
    with pytest.raises(AttributeError) as excinfo:
        dataset.status = 'New Thing'
        dataset.update()
    assert "Dataset.status is read-only." in str(excinfo.value)

def test_tags_is_list_of_strings_only(client, dataset):
    with pytest.raises(AttributeError) as excinfo:
        dataset.tags = 'New Thing'
    assert "Dataset.tags should be a list of strings." in str(excinfo.value)
    with pytest.raises(AttributeError) as excinfo:
        dataset.tags = [1,2,3]
    assert "Dataset.tags should be a list of strings." in str(excinfo.value)
    dataset.tags = ["a","b","c"]
    dataset.update()
    dataset_from_platform = client.get_dataset(dataset.id)
    assert dataset_from_platform.tags == ["a","b","c"]
    
def test_datasets(client, dataset):
    ds_items = len(dataset)

    # create package locally
    pkg = DataPackage('Child of Dataset', package_type='Text')

    assert not pkg.exists
    assert pkg not in dataset

    # add package to dataset
    dataset.add(pkg)

    assert pkg.exists
    assert pkg in dataset
    assert pkg.id in dataset
    assert len(dataset) == ds_items + 1

    # remove from dataset
    dataset.remove(pkg)

    assert not pkg.exists
    assert pkg not in dataset
    assert len(dataset) == ds_items

    # can't create dataset with same name
    with pytest.raises(Exception):
        client.create_dataset(dataset.name)


def test_packages_create_delete(client, dataset):

    # init
    pkg = DataPackage('Some MRI', package_type='MRI')
    assert not pkg.exists

    # create
    dataset.add(pkg)
    assert pkg.exists
    assert pkg.id is not None
    assert pkg.name == 'Some MRI'
    assert pkg.owner_id == client.profile.int_id

    # TODO: (once we auto-include in parent)
    assert pkg in dataset

    # update package name
    pkg.name = 'Some Other MRI'
    pkg = client.update(pkg)

    pkg2 = client.get(pkg.id)

    assert pkg2.name == 'Some Other MRI'
    assert pkg2.id == pkg.id
    assert pkg2.owner_id == client.profile.int_id

    # delete all packages
    client.delete(pkg)

    assert not pkg.exists

    pkg = DataPackage('Something else', package_type='TimeSeries')
    assert not pkg.exists
    dataset.add(pkg)
    assert pkg.exists
    pid = pkg.id
    pkg.delete()
    assert not pkg.exists

    pkg2 = client.get(pid)
    assert pkg2 is None

    # TODO: (once we auto-remove from parent)
    #assert pkg not in dataset

def test_package_type_count(client,dataset):
    n = dataset.package_count()
    pkg = DataPackage('Some MRI', package_type='MRI')
    assert not pkg.exists
    # create
    dataset.add(pkg)
    assert pkg.exists
    client.update(pkg)

    pkg = DataPackage('Something else', package_type='TimeSeries')
    assert not pkg.exists
    dataset.add(pkg)
    assert pkg.exists
    client.update(pkg)

    m = dataset.package_count()
    assert m == n + 2

def test_publish_info(client,dataset):
    publish_info = dataset.published()
    assert publish_info.status =='NOT_PUBLISHED'
    assert publish_info.version_count == 0
    assert publish_info.last_published== None
    assert publish_info.doi == None

def test_owner(client,dataset):
    owner = dataset.owner()
    assert owner.email == client.profile.email

def test_collaborator_user(client,dataset):
    collaborators = dataset.user_collaborators()
    assert len(collaborators) == 1
    assert collaborators[0].email == client.profile.email

def test_collaborator_team(client,dataset):
    collaborators = dataset.team_collaborators()
    assert len(collaborators) == 0

def test_properties(client, dataset):

    pkg = DataPackage('Some Video', package_type='Video')
    assert not pkg.exists

    dataset.add(pkg)
    assert pkg.exists

    pkg.insert_property('my-key','my-value')
    pkg2 = client.get(pkg)
    print('properties =', pkg2.properties)
    assert pkg2.id == pkg.id
    assert pkg2.get_property('my-key').data_type == 'string'
    assert pkg2.get_property('my-key').value == 'my-value'

    explicit_ptypes = {
        'my-int1': ('integer', 123123),
        'my-int2': ('integer', '123123'),
        'my-float': ('double', 123.123),
        'my-float2': ('double', '123.123'),
        'my-float3': ('double', '123123'),
        'my-date': ('date', 1488847449697),
        'my-date2': ('date', 1488847449697.123),
        'my-date3': ('date', datetime.datetime.now()),
        'my-string': ('string', 'my-123123'),
        'my-string2': ('string', '123123'),
        'my-string3': ('string', '123123.123'),
        'my-string4': ('string', 'According to plants, humans are blurry.'),
    }
    for key, (ptype,val) in explicit_ptypes.items():
        pkg.insert_property(key, val, data_type=ptype)
        assert pkg.get_property(key).data_type == ptype

    inferred_ptypes = {
        'my-int1': ('integer', 123123),
        'my-int2': ('integer', '123123'),
        'my-float1': ('double', 123.123),
        'my-float2': ('double', '123.123'),
        'my-date': ('date', datetime.datetime.now()),
        'my-string': ('string', 'i123123'),
        'my-string2': ('string', '#1231'),
    }
    for key, (ptype,val) in inferred_ptypes.items():
        pkg.insert_property(key, val)
        prop = pkg.get_property(key)
        assert prop.data_type == ptype

    # remove property
    pkg.remove_property('my-key')
    assert pkg.get_property('my-key') is None

    pkg2 = client.get(pkg.id)
    assert pkg2.get_property('my-key') is None


def test_can_remove_multiple_items(dataset):
    pkg1 = DataPackage('Some MRI', package_type='MRI')
    dataset.add(pkg1)
    pkg1.update()
    pkg2 = DataPackage('Some Video', package_type='Video')
    dataset.add(pkg2)
    pkg2.update()
    assert pkg1 in dataset.items
    assert pkg2 in dataset.items

    dataset.remove(pkg1)
    dataset.remove(pkg2)
    assert pkg1 not in dataset.items
    assert pkg2 not in dataset.items


def test_timeout():
    with pytest.raises(requests.exceptions.Timeout):
        # initial authentication calls should time out
        get_test_client(max_request_time=0.00001)


def test_client_host_overrides():
    host = 'http://localhost'
    # fails authentication in Blackfynn.__init__
    with pytest.raises(requests.exceptions.RequestException):
        bf = Blackfynn(host=host)

    bf = Blackfynn(model_service_host=host)
    assert bf.settings.model_service_host == host


def test_exception_raise():
    bf = Blackfynn()
    with pytest.raises(Exception) as excinfo:
        bf._api._call('get','/datasets/plop')
    assert "plop not found" in str(excinfo.value)



'''
Description
  The profile management system is used to create multiple profiles for
  different Blackfynn accounts or organizations. The user can also set/unset
  global settings which apply to all profiles.

  Settings are loaded in the following order, with each tier overriding the
  last:

    1. Global settings are loaded first
    2. Default profile, if configured
    3. Command line arguments, e.g.: --profile=<name>
    4. Environment variables, e.g.: BLACKFYNN_API_TOKEN

Usage:
  bf_profile create [<name>]
  bf_profile show [<name>]
  bf_profile delete [<name>] [-f | --force]
  bf_profile list [-c | --contents]
  bf_profile set-default [<name>]
  bf_profile unset-default [-f | --force]
  bf_profile status
  bf_profile set <key> <value> [-f | --force] [--profile=<name>]
  bf_profile unset <key> [-f | --force] [--profile=<name>]
  bf_profile keys [--profile=<name>]
  bf_profile version
  bf_profile help

Options:
  -c --contents       List profiles with contents. Also lists global settings, if any
  -f --force          Attempt action without prompting for confirmation. Use with care
  --profile=<name>    Use specified profile (instead of default)
  -h --help           Show help

Advanced commands:
  set (global|<name>) <key> <value>   Set key/value pair for given profile, or globally
  unset (global|<name>) <key>         Unset key/value pair for given profile, or globally
  keys                                List all available keys and their default values

For additional features, install the Blackfynn CLI Agent:
https://developer.blackfynn.io/agent
'''

from __future__ import absolute_import, print_function
import io
import os
from builtins import input

from docopt import docopt

import blackfynn
from blackfynn import Blackfynn, DEFAULT_SETTINGS, Settings


def main():
    args = docopt(__doc__, version=blackfynn.__version__)

    # Test for these two commands first as they
    # do not require a Blackfynn client or reading the settings file
    if args['help']:
        print(__doc__.strip('\n'))
        return
    elif args['version']:
        print(blackfynn.__version__)
        return

    settings = Settings(args['--profile'])
    if not os.path.exists(settings.config_file):
        setup_assistant(settings)
    elif args['create']:
        create_profile(settings, args['<name>'])
    elif args['show']:
        show_profile(settings, args['<name>'])
    elif args['delete']:
        delete_profile(settings, args['<name>'], args['--force'])
    elif args['list']:
        list_profiles(settings, args['--contents'])
    elif args['set-default']:
        set_default(settings, args['<name>'])
    elif args['unset-default']:
        unset_default(settings, args['--force'])
    elif args['set']:
        set_key(settings, args['<key>'], args['<value>'], args['--profile'], args['--force'])
    elif args['unset']:
        unset_key(settings, args['<key>'], args['--profile'], args['--force'])
    elif args['keys']:
        list_keys(settings, args['--profile'])
    elif args['status']:
        bf = Blackfynn(args['--profile'])
        show_status(bf)
    else:
        invalid_usage()

    with io.open(settings.config_file, 'w') as configfile:
        settings.config.write(configfile)


def setup_assistant(settings):
    settings.config.clear()

    print("Blackfynn profile setup assistant")

    settings.config['global'] = {'default_profile' : 'none'}

    print("Create a profile:")
    create_profile(settings)

    print("Setup complete. Run 'bf_profile help' for available commands and actions")


# User commands
#=======================================================

def create_profile(settings, name=None):
    if not name:
        name = input('  Profile name [default]: ') or 'default'

    if name in settings.config:
        if name in ('global', 'agent', 'none'):
            print("Profile name '{}' reserved for system. Please try a different name".format(name))
        else:
            print("Profile '{}' already exists".format(name))
        abort()

    print("Creating profile '{}'".format(name))
    settings.config[name] = {}

    settings.config[name]['api_token']  = input('  API token: ')
    settings.config[name]['api_secret'] = input('  API secret: ')

    if settings.config['global']['default_profile'] == 'none':
        settings.config['global']['default_profile'] = name
        print("Default profile: {}".format(name))

    else:
        if yesno_prompt("Would you like to set '{}' as default (Y/n)? ".format(name)):
            set_default(settings, [name])


def delete_profile(settings, name, force):
    if not name:
        name = input('  Profile to delete: ')

    if not valid_name(settings, name):
        abort()

    if not force:
        if not yesno_prompt("Delete profile '{}' (Y/n)? ".format(name)):
            print('abort')
            abort()

    print("Deleting profile '{}'".format(name))
    settings.config.remove_section(name)

    if settings.config['global']['default_profile'] == name:
        settings.config['global']['default_profile'] = 'none'
        print("\033[31m* Warning: default profile unset. Use 'bf profile set-default <name>' to set a new default\033[0m")


def list_profiles(settings, contents):
    '''
    Lists all profiles
    '''
    print('Profiles:')
    for section in settings.config.sections():
        if section not in ['global']:
            if section == settings.config['global']['default_profile']:
                print('* \033[32m{}\033[0m'.format(section))
            else:
                print('  {}'.format(section))
            if contents:
                print_profile(settings, section, 4)
    if contents:
        if len(settings.config['global']) > 1:
            print('Global Settings:')
            print_profile(settings, 'global', 2)


def set_default(settings, name):
    if not name:
        name = input('  Profile name to set as default: ')

    if not valid_name(settings, name):
        abort()

    print("Default profile: {}".format(name))
    settings.config['global']['default_profile'] = name


def unset_default(settings, force):
    original = settings.config['global']['default_profile']

    if not force:
        if not yesno_prompt("Unset default profile '{}' (Y/n)? ".format(original)):
            print('abort')
            abort()

    print("Default profile '{}' unset. Using global settings and environment variables".format(original))
    settings.config['global']['default_profile'] = 'none'


def show_profile(settings, profile):
    if profile:
        name = profile
    else:
        name = settings.default_profile

    if name == 'global' or valid_name(settings, name):
        print('{} contents:'.format(name))
        print_profile(settings, name, 2, True)


# Advanced commands
#=======================================================

def set_key(settings, key, value, profile, force):
    if not key in DEFAULT_SETTINGS:
        print("Invalid key: '{}'\n see 'bf profile keys' for available keys".format(key))
        return

    if profile:
        name = profile
    else:
        name = settings.default_profile

    if not name in settings.config:
        print("Profile '{}' does not exist".format(name))
        return

    if not force and key in settings.config[name]:
        if not yesno_prompt("{}: {} already set. Overwrite (Y/n)? ".format(name, key)):
            abort()

    print("{}: {}={}".format(name, key, value))
    settings.config[name][key] = value


def unset_key(settings, key, profile, force):
    if profile:
        name = profile
    else:
        name = settings.default_profile

    if not name in settings.config:
        print("Profile '{}' does not exist".format(name))
        abort()

    if not key in settings.config[name]:
        print("{}: {} not set".format(name, key))
        abort()

    if not force and key in settings.config[name]:
        if not yesno_prompt("{}: Unset {} (Y/n)? ".format(name, key)):
            abort()

    print("{}: {} unset".format(name, key))
    settings.config[name].pop(key)


def list_keys(settings, profile):
    if profile:
        name = profile
    else:
        name = settings.default_profile

    print("Keys and default values for '{}'".format(name))
    for key, value in sorted(settings.profiles[name].items()):
        print('  {} : {}'.format(key, value))


def show_status(bf):
    print('Active profile:\n  \033[32m{}\033[0m\n'.format(bf.settings.active_profile))

    if bf.settings.env:
        key_len = 0
        value_len = 0
        for key, value in bf.settings.env.items():
            valstr = '{}'.format(value)
            key_len = max(key_len, len(key))
            value_len = max(value_len, len(valstr))

        print('Environment variables:')
        print('  \033[4m{:{key_len}}    {:{value_len}}    {}'.format(
            'Key', 'Value', 'Environment Variable\033[0m', key_len=key_len, value_len=value_len))
        for value, evar in sorted(bf.settings.env.items()):
            valstr = '{}'.format(value)
            # get internal variable
            ivar = ''
            for k, v in bf.settings.__dict__.items():
                if str(v) == str(evar):
                    ivar = k
            print('  {:{key_len}}    {:{value_len}}    {}'.format(
                ivar, evar, valstr, key_len=key_len, value_len=value_len))
        print()

    print("Blackfynn environment:")
    print("  User               : {}".format(bf.profile.email))
    print("  Organization       : {} (id: {})".format(bf.context.name, bf.context.id))
    print("  API Location       : {}".format(bf.settings.api_host))


#Helper functions
#=======================================================

def abort():
    exit(1)


def yesno_prompt(msg):
    return input(msg).lower() in ('y', 'yes')


def invalid_usage():
    print('Invalid usage. See `bf_profile help` for available commands')
    abort()


def valid_name(settings, name):
    if name not in settings.config or name == 'global':
        print("Profile '{}' does not exist".format(name))
        return False
    return True


def print_profile(settings, name, indent=0, show_all=False):
    if show_all:
        key_len = 0
        for key in settings.profiles[name].keys():
            key_len = max(key_len, len(key))

        for key, value in sorted(settings.profiles[name].items()):
            if key != 'default_profile':
                if key in settings.config[name] and name != 'global':
                    print(' '*indent + '{:{key_len}} : \033[32m{}\033[0m ({})'.format(key, value, name, key_len=key_len))
                elif key in settings.config['global']:
                    print(' '*indent + '{:{key_len}} : \033[34m{}\033[0m (global)'.format(key, value, key_len=key_len))
                else:
                    print(' '*indent + '{:{key_len}} : \033[0m{}\033[0m'.format(key, value, key_len=key_len))

    else:
        key_len = 0
        for key in settings.config[name].keys():
            key_len = max(key_len, len(key))

        for key, value in sorted(settings.config[name].items()):
            if key != 'default_profile':
                print(' '*indent + '{:{key_len}} : {}'.format(key, value, key_len=key_len))



from __future__ import absolute_import, division, print_function

import logging
import os

logging.basicConfig()


def get_logger(name='blackfynn-python'):
    """
    Returns a logger configured to be used throughout the
    blackfynn-python library

    Args:
      name (string): Name of the logger, default "blackfynn-python"

    Returns:
      Logger
    """

    logger = logging.getLogger(name)
    logger.setLevel(os.environ.get('BLACKFYNN_LOG_LEVEL', 'INFO'))

    return logger



from __future__ import absolute_import, division, print_function

import io
import re
from os import path

from setuptools import find_packages, setup

with io.open('blackfynn/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

here = path.abspath(path.dirname(__file__))

with io.open(path.join(here, 'requirements.txt'), mode='r', encoding='utf-8') as f:
    reqs = [line.strip() for line in f if not line.startswith('#')]

# Get the long description from the README file
with io.open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name = "blackfynn",
    version = version,
    author = "Blackfynn, Inc.",
    author_email = "mark@blackfynn.com",
    description = "Python client for the Blackfynn Platform",
    long_description = long_description,
    packages=find_packages(),
    package_dir={'blackfynn': 'blackfynn'},
    setup_requires=['cython'],
    install_requires = reqs,
    extras_require={
        "data":  ["numpy>=1.13", "pandas>=0.20"],
    },
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4.0',
    entry_points = {
        'console_scripts': [
            'bf_profile=blackfynn.cli.bf_profile:main',
        ]
    },
    license = "",
    keywords = "blackfynn client rest api",
    url = "https://github.com/Blackfynn/blackfynn-python",
    project_urls = {
        "Blackfynn": "https://www.blackfynn.com",
        # Do not remove or rename this "Documentation" URL: it is used by
        # the Sphinx theme to generate links to previous doc versions.
        # However, you can change the URL with no issues.
        "Documentation": "https://developer.blackfynn.io/python",
        "Bug Reports": "https://github.com/Blackfynn/blackfynn-python/issues"
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ]
)



# -*- coding: utf-8 -*-

# blackfynn-specific
from __future__ import absolute_import, division, print_function
from builtins import dict, object
from future.utils import as_native_str

import blackfynn.log as log
from blackfynn import Settings
from blackfynn.api.concepts import (
    ModelRelationshipInstancesAPI,
    ModelRelationshipsAPI,
    ModelsAPI,
    ModelTemplatesAPI,
    RecordsAPI
)
from blackfynn.api.core import (
    CoreAPI,
    OrganizationsAPI,
    SearchAPI,
    SecurityAPI
)
from blackfynn.api.data import (
    DataAPI,
    DatasetsAPI,
    FilesAPI,
    PackagesAPI,
    TabularAPI
)
from blackfynn.api.timeseries import TimeSeriesAPI
from blackfynn.api.transfers import IOAPI
from blackfynn.api.user import UserAPI
from blackfynn.base import ClientSession
from blackfynn.models import Dataset, ModelTemplate


class Blackfynn(object):
    """
    The client provides an interface to the Blackfynn platform, giving you the
    ability to retrieve, add, and manipulate data.

    Args:
        profile (str, optional): Preferred profile to use
        api_token (str, optional): Preferred api token to use
        api_secret (str, optional): Preferred api secret to use
        model_service_host (str, optional): Preferred model service host to use
        host (str, optional): Preferred host to use
        env_override (bool, optional): Should environment variables override settings
        **overrides (dict, optional): Settings to override

    Examples:
        Load the client library and initialize::

            from blackfynn import Blackfynn

            # note: no API token needed if environment variables are set
            bf = Blackfynn()

        Retrieve/modify items on the platform::

            # get some dataset
            ds = bf.get_dataset('my dataset')
            print("Dataset {} has ID={}".format(ds.name, ds.id))

            # list all first-level items in dataset
            for item in ds:
                print("Item: {}".format(item))

            # grab some data package
            pkg = bf.get("N:package:1234-1234-1234-1234")

            # modify a package's name
            pkg.name = "New Package Name"

            # add some property
            pkg.set_property('Room', 'ICU-123')

            # sync changes
            pkg.update()


    Note:
        To initialize your ``Blackfynn`` client without passing any arguments,
        ensure that your ``BLACKFYNN_API_TOKEN`` and ``BLACKFYNN_API_SECRET`` environment variables
        are properly set.

    """
    def __init__(self, profile=None, api_token=None, api_secret=None, jwt=None, host=None, model_service_host=None, env_override=True, **overrides):

        self._logger = log.get_logger("blackfynn.client.Blackfynn")

        overrides.update({ k: v for k, v in {
            'api_token': api_token,
            'api_secret': api_secret,
            'api_host': host,
            'jwt': jwt,
            'model_service_host': model_service_host
            }.items() if v != None })
        self.settings = Settings(profile, overrides, env_override)

        if (self.settings.api_token is None or self.settings.api_secret is None) and self.settings.jwt is None:
            if self.settings.api_token  is None:
                raise Exception('Error: No API token found. Cannot connect to Blackfynn.')
            if self.settings.api_secret is None:
                raise Exception('Error: No API secret found. Cannot connect to Blackfynn.')
            if self.settings.jwt  is None:
                raise Exception('Error: No JWT found. Cannot connect to Blackfynn.')

        # direct interface to REST API.
        self._api = ClientSession(self.settings)

        # account
        self._api.authenticate()

        self._api.register(
            CoreAPI,
            OrganizationsAPI,
            DatasetsAPI,
            FilesAPI,
            DataAPI,
            PackagesAPI,
            TimeSeriesAPI,
            TabularAPI,
            SecurityAPI,
            SearchAPI,
            IOAPI,
            UserAPI,
            ModelsAPI,
            RecordsAPI,
            ModelRelationshipsAPI,
            ModelRelationshipInstancesAPI,
            ModelTemplatesAPI
        )

        self._api._context = self._api.organizations.get(self._api._organization)


    @property
    def context(self):
        """
        The current organizational context of the active client.
        """
        return self._api._context

    @property
    def profile(self):
        """
        The profile of the current active user.
        """
        return self._api.profile

    def organizations(self):
        """
        Return all organizations for user.
        """
        return self._api.organizations.get_all()

    def datasets(self):
        """
        Return all datasets for user for an organization (current context).
        """
        self._check_context()
        return self.context.datasets

    def get(self, id, update=True):
        """
        Get any DataPackage or Collection object by ID.

        Args:
            id (str): The ID of the Blackfynn object.

        Returns:
            Object of type/subtype ``DataPackage`` or ``Collection``.
        """
        try:
            return self._api.core.get(id, update=update)
        except:
            self._logger.info("Unable to retrieve object"\
                             "\n\nAcceptable objects for get() are:"\
                             "\n - DataPackages"\
                             "\n - Collections"\
                             "\n\nUse get_dataset() if trying to retrieve a dataset")

    def create(self, thing):
        """
        Create a object on the platform.
        """
        return self._api.core.create(thing)

    def create_dataset(self, name, description=None, automatically_process_packages=False):
        """
        Create a dataset under the active organization.

        Args:
            name (str): The name of the to-be-created dataset

        Returns:
            The created ``Dataset`` object

        """
        self._check_context()
        return self._api.datasets.create(Dataset(name, description=description, automatically_process_packages=automatically_process_packages))

    def get_dataset(self, name_or_id):
        """
        Get Dataset by name or ID.

        Args:
            name_or_id (str): the name or the ID of the dataset

        Note:
            When using name, this method gnores case, spaces, hyphens,
            and underscores such that these are equivelent:

              - "My Dataset"
              - "My-dataset"
              - "mydataset"
              - "my_DataSet"
              - "mYdata SET"

        """
        try:
            return self._api.datasets.get(name_or_id)
        except:
            pass

        result = self._api.datasets.get_by_name_or_id(name_or_id)
        if result is None:
            raise Exception("No dataset matching name or ID '{}'.".format(name_or_id))
        return result

    def update(self, thing):
        """
        Update an item on the platform.

        Args:
            thing (object or str): the ID or object to update

        Example::

            my_eeg = bf.get('N:package:1234-1234-1234-1234')
            my_eeg.name = "New EEG Name"
            bf.update(my_eeg)

        Note:
            You can also call update directly on objects, for example::

                my_eeg.name = "New EEG Name"
                my_eeg.update()

        """
        return self._api.core.update(thing)

    def delete(self, *things):
        """
        Deletes items from the platform.

        Args:
            things (list): ID or object of items to delete

        Example::

            # delete a package by passing in object
            bf.delete(my_package)

            # delete a package by ID
            bf.delete('N:package:1234-1234-1234-1235')

            # delete many things (mix of IDs and objects)
            bf.delete(my_package, 'N:collection:1234-1234-1234-1235', another_collection)

        Note:
            When deleting ``Collection``, all child items will also be deleted.

        """
        return self._api.core.delete(*things)

    def move(self, destination, *things):
        """
        Moves item(s) from their current location to the destination.

        Args:
            destination (object or str): The ID of the destination. This must be of type
                type ``Collection`` or ``None``. If destination is ``None``, ``things``
                will be moved to the top of their containing ``Dataset``
            things (list): the IDs or objects to move.

        """
        r = self._api.data.move(destination, *things)

    def search(self, query, max_results=10):
        """
        Find an object on the platform.

        Args:
            query (str): query string to perform search.
            max_results (int, optional): the number of results to return

        Example::

            # find some items belonging to patient123
            for result in  bf.search('patient123'):
                print("found:", result)

        """
        return self._api.search.query(query, max_results=max_results)

    def _check_context(self):
        if self.context is None:
            raise Exception('Must set context before executing method.')


    def get_model_template(self, template_id):
        """
        Get a model template belonging to the current organization.

        Args:
            template_id (str): The ID of the model template object.

        Returns:
            Object of type ModelTemplate.
        """
        return self._api.templates.get(template_id)

    def get_model_templates(self):
        """
        Get all of the model templates belonging to the current organization.

        Returns:
            A list of objects of type ModelTemplate.
        """
        return self._api.templates.get_all()

    def create_model_template(
        self,
        name=None,
        description=None,
        category=None,
        required=None,
        properties=None,
        template=None
    ):
        """
        Create a new model template belonging to the current organization.

        Args:
            template (ModelTemplate): A ModelTemplate object.
            OR
            name (str): The name of the template
            description (str, optional): A description of the template
            category (str): The category of the template
            required ([str], optional): The names of the required template properties
            properties (dict OR tuples): The template properties

        Example::

            my_template = ModelTemplate(
                schema="http://schema.blackfynn.io/model/draft-01/schema",
                name="Patient",
                description="This is a patient.",
                category="Person",
                properties=dict(
                    name=dict(
                        type="string",
                        description="Name"
                    ),
                    DOB=dict(
                        type="string",
                        format="date",
                        description="Date of Birth"
                    )
                ),
                required=["name"]
            bf.create_model_template(my_template)

        Example::

            bf.create_model_template(
                name="Patient",
                description="This is a patient.",
                category="Person",
                required=["name"],
                properties=[("name", "string"), ("DOB", "date")]
            )

        Returns:
            The created ModelTemplate object.
        """
        if (template is None):
            template_to_create = ModelTemplate(name=name, description=description, category=category, required=required,
                                     properties=properties)
        else:
            template_to_create = template
            assert isinstance(template_to_create, ModelTemplate), "template must be type ModelTemplate"

        return self._api.templates.create(template=template_to_create)

    def validate_model_template(
        self,
        name=None,
        description=None,
        category=None,
        required=None,
        properties=None,
        template=None
    ):
        """
        Validate a model template schema.

        Args:
            template (ModelTemplate): A ModelTemplate object.
            OR
            name (str): The name of the template
            description (str, optional): A description of the template
            category (str): The category of the template
            required ([str], optional): The names of the required template properties
            properties (dict OR tuples): The template properties

        Example::

            my_template = ModelTemplate(
                schema="http://schema.blackfynn.io/model/draft-01/schema",
                name="Patient",
                description="This is a patient.",
                category="Person",
                properties=dict(
                    name=dict(
                        type="string",
                        description="Name"
                    ),
                    DOB=dict(
                        type="string",
                        format="date",
                        description="Date of Birth"
                    )
                ),
                required=["name"]
            bf.validate_model_template(my_template)

        Example::

            bf.validate_model_template(
                name="Patient",
                category="Person",
                properties=[("name", "string"), ("Weight", "number")]
            )

        Returns:
            "True" or a list of validation errors
        """
        if (template is None):
            template_to_validate = ModelTemplate(name=name, description=description, category=category, required=required,
                                               properties=properties)
        else:
            template_to_validate = template
            assert isinstance(template_to_validate, ModelTemplate), "template must be type ModelTemplate"

        return self._api.templates.validate(template=template_to_validate)

    def delete_model_template(self, template_id):
        """
        Deletes a model template from the platform.

        Args:
            template_id (str): The ID of the model template object.

        """
        return self._api.templates.delete(template_id)

    @as_native_str()
    def __repr__(self):
        return "<Blackfynn user='{}' organization='{}'>".format(self.profile.email, self.context.name)



import pytest
from uuid import uuid4

from blackfynn.models import (
    LinkedModelProperty,
    ModelProperty,
)
from tests.utils import create_test_dataset

### Testing linked properties locally:
def test_make_linked_property(dataset):
    # make a new model
    model = dataset.create_model('my_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])

    # create a linked property linking to that model,
    # and make sure the link initialized correctly
    link1 = LinkedModelProperty('link1', model, '1st linked property')

    dct1 = link1.as_dict()
    assert dct1['name'] == 'link1'
    assert dct1['displayName'] == '1st linked property'
    assert dct1['to'] == model.id

    # make a linked property from a dict,
    # and make sure it initialized correctly
    link2 = LinkedModelProperty.from_dict({'link': {
        'name': 'link2',
        'displayName': '2nd linked property',
        'to': model.id,
        'id': 'XXX-XXX-XXX',
        'position': 0}})
    assert link2.id == 'XXX-XXX-XXX'
    assert link2.position == 0

    dct2 = link2.as_dict()
    assert dct2['name'] == 'link2'
    assert dct2['displayName'] == '2nd linked property'
    assert dct2['to'] == model.id


### Testing linked property API methods:
def test_add_linked_property(dataset):
    # Create two models and link one to the other
    source = dataset.create_model('source_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])
    target = dataset.create_model('target_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])
    source.add_linked_property('link', target, 'my linked property')

    # Make sure newly created link is accessible through the API
    assert any(l.name == 'link' and l.target == target.id for l in dataset.get_topology()['linked_properties'])
    assert 'link' in source.linked

    # Prevent user from adding duplicate linked properties
    with pytest.raises(Exception):
        source.add_linked_property('link', source, 'duplicate linked property')
    assert not any(l.display_name == 'duplicate linked property' for l in dataset.get_topology()['linked_properties'])
    
def test_add_linked_property_bulk(dataset):
    # Link one model to three others
    source = dataset.create_model('source_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])
    target = dataset.create_model('target_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])
    link1 = LinkedModelProperty('link1', target, 'bulk-added')
    link2 = LinkedModelProperty('link2', target, 'bulk-added')
    link3 = LinkedModelProperty('link3', target, 'bulk-added')
    source.add_linked_properties([link1, link2, link3])

    # Make sure newly created links are accessible through the API
    bulk = [x for x in dataset.get_topology()['linked_properties'] if x.display_name == 'bulk-added']
    assert len(bulk) == 3
    assert len(source.linked) == 3

def test_edit_linked_property(dataset):
    # make a model and add a linked property
    source = dataset.create_model('source_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])
    target = dataset.create_model('target_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])
    link = source.add_linked_property('link', target, 'my linked property')

    # edit the linked property and update model
    link.display_name = 'updated linked property'
    link.position = 99
    source.update()
    
    # Make sure changes were saved
    new_link = source.get_linked_property('link')
    assert new_link.position == 99
    assert new_link.display_name == 'updated linked property'

def test_delete_linked_property(dataset):
    # make a model and add a linked property
    source = dataset.create_model('source_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])
    target = dataset.create_model('target_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])
    source.add_linked_property('link', target, 'my linked property')

    # delete the link
    source.remove_linked_property('link')

    # Make sure changes were saved
    assert 'link' not in source.linked

def test_retrieve_linked_properties(dataset):
    # make a model and add a linked property
    source = dataset.create_model('source_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])
    target = dataset.create_model('target_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])
    source.add_linked_property('link', target, 'my linked property')

    new_source = dataset.get_model(source.type)
    assert 'link' in new_source.linked


### Testing linked property values
def test_add_link(dataset):
    # make a model and add a linked property
    source = dataset.create_model('source_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])
    target = dataset.create_model('target_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])
    prop = source.add_linked_property('link', target, 'my linked property')

    # make records and link them
    source_rec = source.create_record({'name': 'source_record'})
    target_rec = target.create_record({'name': 'target_record'})
    link = source_rec.add_linked_value(target_rec, prop)
    assert link.id == source_rec.get_linked_value('link').id

    # prevent duplicate links from being created
    target_rec2 = target.create_record({'name': 'second_target'})
    link2 = source_rec.add_linked_value(target_rec2, prop)
    links = source_rec.get_linked_values()
    assert len(links) == 1
    assert links[0].target_record_id == target_rec2.id

def test_get_link(dataset):
    # make a model and add a linked property
    source = dataset.create_model('source_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])
    target = dataset.create_model('target_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])
    prop = source.add_linked_property('link', target, 'my linked property')

    # make records and link them
    source_rec = source.create_record({'name': 'source_record'})
    target_rec = target.create_record({'name': 'target_record'})
    link = source_rec.add_linked_value(target_rec, prop)
    link2 = source_rec.get_linked_value('link')
    link3 = source_rec.get_linked_value(link.id)
    assert link.id == link2.id == link3.id
    assert link.source_model.id == source.id
    assert link.target_model.id == target.id

def test_remove_link(dataset):
    # make a model and add a linked property
    source = dataset.create_model('source_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])
    target = dataset.create_model('target_model_{}'.format(uuid4()), schema=[ModelProperty('name', title=True)])
    prop = source.add_linked_property('link', target, 'my linked property')

    # make records and link them
    source_rec = source.create_record({'name': 'source_record'})
    target_rec = target.create_record({'name': 'target_record'})
    link = source_rec.add_linked_value(target_rec, prop)

    # delete the link
    source_rec.delete_linked_value(link.id)
    assert not any(link.target == target_rec.id for link in source_rec.get_linked_values())



from blackfynn.models import _get_all_class_args
from builtins import object

def test_get_all_class_args():
    class A(object):
        def __init__(self, x, y, **kwargs):
            pass

    class B(A):
        def __init__(self, y, z, *args):
            pass

    assert _get_all_class_args(B) == set(['self', 'x', 'y', 'z', 'args', 'kwargs'])
